// bagg.h: header for the Bayesian Action Graph Game class

#ifndef __BAGG_H
#define __BAGG_H



#include <iostream>
#include "agg.h"



#ifdef USE_CVECTOR
#include "../cmatrix.h"
typedef cvector ProbDist;
//typedef cvector NumberVector;
#else
typedef  vector<Number>  ProbDist;
//typedef vector<Number> NumberVector;
#endif

using namespace std;


class bagg {

public:
  static const char COMMENT_CHAR= '#';

  friend class agentform;

  friend ostream& operator<< (ostream& s, const bagg& g);

  static bagg* makeBAGG(char *filename);
  static bagg* makeBAGG(istream& in);
  static bagg* makeRandomBAGG(int n,vector<int> &numTypes,vector<ProbDist> &TDist,int S,int P,
	vector<vector<vector<int> > > &TASets, 
	vector<vector<int> > &neighb, 
	vector<projtype> &projTypes,
	int seed, bool int_payoffs=false, int int_factor=100);

  bagg(int N, int S, 
    vector<int>& numTypes,
    vector<ProbDist>& TDist,
    vector<vector<vector<int > > > &typeActionSets,
    vector<vector<vector<int > > > &ta2a,
    agg* aggPtr);

  ~bagg(){
    delete[] typeOffset;
    delete[] strategyOffset;
    delete aggPtr;
  }

  inline int getNumPlayers() const {return numPlayers;}
  inline int getNumTypes() const {return typeOffset[numPlayers];}
  inline int getNumTypes(int i) const {return numTypes[i];}
  inline int getNumActions(int pl, int t) const {return typeActionSets[pl][t].size();}

  inline int getNumActionNodes() const {return numActionNodes;}
  inline int getNumFunctionNodes() const {return aggPtr->getNumFunctionNodes();}

  inline int firstAction(int pl,int t) {return strategyOffset[typeOffset[pl]+t];}
  inline int lastAction(int pl, int t) {return strategyOffset[typeOffset[pl]+t+1];}

  //exp. payoff under mixed strat profile
  Number getMixedPayoff(int player, StrategyProfile &s);
  //exp payoff for player, conditioned on her receiving type tp.
  Number getMixedPayoff(int player,int tp, StrategyProfile &s);

  void getPayoffVector(NumberVector &dest, int player,int tp, const StrategyProfile &s);
  Number getV (int player, int tp, int action,const StrategyProfile &s);

  Number getPurePayoff(int player, int tp, int *s);
  Number getPurePayoff(int player, int *s){
    Number r=0;
    for(int i=0;i<numTypes[player];++i){
      r+= indepTypeDist[player][i] * getPurePayoff(player, i, s);
    }
    return r;
  }

  Number getSymMixedPayoff(StrategyProfile &s);

  Number getSymMixedPayoff(int tp, StrategyProfile &s);

  Number getSymMixedPayoff(int tp, int act, StrategyProfile &s);

private:

  int numPlayers;
  int numActionNodes;
  vector<int> numTypes;
  vector<ProbDist> indepTypeDist;


  //for each player, each of her types, the set of actions
  vector<vector<vector<int> > > typeActionSets;


  vector<vector<vector<int > > > typeAction2ActionIndex;

  int *typeOffset;
  int *strategyOffset;
  

  //AGG that stores the action graph and utility functions
  agg* aggPtr;



  static void stripComment(istream& in);

  void getAGGStrat(StrategyProfile &as, const StrategyProfile &s, int player=-1, int tp=-1, int action=-1);
  void getSymAGGStrat(StrategyProfile &as, const StrategyProfile &s);

};

ostream& operator<< (ostream& s, const bagg& g);


#endif
