#ifndef __AGENTFORM_H
#define __AGENTFORM_H

//agent form wrapper for bayesian agg



#include "cmatrix.h"
#include "gnmgame.h"
#include "BAGG/bagg.h"

using namespace std;


class agentform : public gnmgame {
public:

  //static agentform* makeAgentForm(char* filename){ return new agentform(bagg::makeBAGG(filename));}
  //static agentform* makeAgentForm(istream& in){ return new agentform(bagg::makeBAGG(in)); }

  agentform(bagg* ptr)
    :gnmgame(ptr->getNumTypes(), makeActions(ptr) ), baggPtr(ptr),agent2baggPlayer(ptr->getNumTypes())
  {
    int k=0;
    for (int i=0;i<ptr->getNumPlayers();i++){
      for (int j=0;j<ptr->getNumTypes(i);j++,k++){
        agent2baggPlayer[k] = i;
      }
    }
  }

  ~agentform()
  {
    //delete baggPtr;
  }

  int getNumActionNodes() {return baggPtr->getNumActionNodes();}

  Number getMixedPayoff(int player, cvector &s){
      int p = agent2baggPlayer[player];
      int tp = player - baggPtr->typeOffset[p];
      return baggPtr->getMixedPayoff(p, tp, s);
  }

  Number getPurePayoff(int player, int* s){
      int p = agent2baggPlayer[player];
      int tp= player - baggPtr->typeOffset[p];
      return baggPtr->getPurePayoff(p,tp,s);
  }

  void payoffMatrix(cmatrix &dest, cvector &s, Number fuzz){
    Number fuzzcount;
    vector<int> done(baggPtr->aggPtr->getNumActions(), -1);
    cmatrix d (baggPtr->aggPtr->getNumActions(), baggPtr->aggPtr->getNumActions());
    cvector as (baggPtr->aggPtr->getNumActions());
    baggPtr->getAGGStrat(as, s); //get the induced AGG strategies
    baggPtr->aggPtr->payoffMatrix(d,as,fuzz);

    //diagonals
    for (int rown=0; rown<getNumPlayers(); ++rown){
      fuzzcount=fuzz;
      for (int rowi=firstAction(rown); rowi<lastAction(rown);rowi++){
        for (int coli=firstAction(rown);coli<lastAction(rown);coli++){
  		  dest[rowi][coli]=fuzzcount;
  		  fuzzcount+=fuzz;
        }
      }
    }

    for (int rowp =0; rowp<getNumPlayers();rowp++){//row player of the agent form
      int rowbp=agent2baggPlayer[rowp];
      int rowtp = rowp - baggPtr->typeOffset[rowbp];
      for (int rowa = 0; rowa<getNumActions(rowp);rowa++){//row action of agent form
        //get corresponding action in the agg
        int rowaact = baggPtr->typeAction2ActionIndex[rowbp][rowtp][rowa]+baggPtr->aggPtr->firstAction(rowbp);
        


        if (done.at(rowaact) != -1){//this row have been done earlier, copy
            for(int colp=0;colp<getNumPlayers(); ++colp)if(colp!=rowp){
                int colbp=agent2baggPlayer[colp];
                if (colbp==rowbp){
                    //here the values being copied are idential;
                    //just need to avoid copying from the diagonal block of row done[rowaact].
                    for(int coli=firstAction(colp);coli<lastAction(colp);++coli)
                        dest[rowa+firstAction(rowp)][coli]=dest[done[rowaact]][firstAction(rowp)];
                }
                else{
                    for(int coli=firstAction(colp);coli<lastAction(colp);++coli)
                        dest[rowa+firstAction(rowp)][coli]=dest[done[rowaact]][coli]; //straight copy
                }
            }
            continue; //go do next row
        }

        done.at(rowaact)=rowa+firstAction(rowp);

        for (int colbp=0; colbp<baggPtr->getNumPlayers(); colbp++){//col player of Bayesian game
          
          if (colbp==rowbp){
            if (baggPtr->getNumTypes(colbp)>1) {
              //just need to calculate one value, the EU under deviation to rowaact of the AGG
              int colbp2=0; //a player of the BAGG that is not colbp
              for (;colbp2<baggPtr->getNumPlayers();colbp2++){
                if (colbp2!=colbp) break;
              }
              Number u=0;
              for (int aact=baggPtr->aggPtr->firstAction(colbp2); aact<baggPtr->aggPtr->lastAction(colbp2);++aact){
                  u+=d[rowaact][aact]*as[aact];
              }

              for (int colp=baggPtr->typeOffset[colbp];colp<baggPtr->typeOffset[colbp+1];colp++)if(colp!=rowp){
                  for (int coli=firstAction(colp);coli<lastAction(colp);++coli) dest[rowa+firstAction(rowp)][coli]=u;
              }
            }
          }
          else{
            for (int coltp=0;coltp<baggPtr->getNumTypes(colbp);++coltp){//col type of Bayesian game
              int colp = baggPtr->typeOffset[colbp] + coltp;
              //if (colp==rowp) continue;
              //get induced AGG strategy for colbp, except contribution from type coltp:
              for (size_t act=0; act< baggPtr->typeActionSets[colbp][coltp].size(); ++act){
                  int aact= baggPtr->typeAction2ActionIndex[colbp][coltp][act];
                  as[aact+baggPtr->aggPtr->firstAction(colbp)] -= baggPtr->indepTypeDist[colbp][coltp] * s[act+firstAction(colp)];
              }
              Number u=0;
              for (int aact=baggPtr->aggPtr->firstAction(colbp); aact<baggPtr->aggPtr->lastAction(colbp) ; ++aact){
                  u+=d[rowaact][aact]*as[aact];
              }
              for (int cola=0; cola<getNumActions(colp); cola++){//col action of agent form
                int colaact=baggPtr->typeAction2ActionIndex[colbp][coltp][cola]+baggPtr->aggPtr->firstAction(colbp);
                //as[colaact] += baggPtr->indepTypeDist[colbp][coltp];
                dest[rowa+firstAction(rowp)][cola+firstAction(colp)] =
                    u +
                    d[rowaact][colaact]*baggPtr->indepTypeDist[colbp][coltp];

                //for (int aact=baggPtr->aggPtr->firstAction(colbp); aact<baggPtr->aggPtr->lastAction(colbp) ; ++aact){
                //    dest[rowa+firstAction(rowp)][cola+firstAction(colp)] += d[rowaact][aact] * as[aact];
                //}
                //as[colaact] -= baggPtr->indepTypeDist[colbp][coltp];
              }
              //restore as to original values
              for (size_t act=0; act< baggPtr->typeActionSets[colbp][coltp].size(); ++act){
                  int aact= baggPtr->typeAction2ActionIndex[colbp][coltp][act];
                  as[aact+baggPtr->aggPtr->firstAction(colbp)] += baggPtr->indepTypeDist[colbp][coltp] * s[act+firstAction(colp)];
              }

            }//end for coltp
          }//end else
        }
      }
    }
  }

  void getPayoffVector(cvector &dest, int player,const cvector &s){
      int p = agent2baggPlayer[player];
      int tp = player - baggPtr->typeOffset[p];
      baggPtr->getPayoffVector(dest, p, tp, s);
  }


//private:
  bagg* baggPtr;

  vector<int> agent2baggPlayer;
private:

  static int* makeActions(bagg* baggPtr){
    int* actions=new int[baggPtr->getNumTypes()];
    int k=0;
    for (int i=0;i<baggPtr->getNumPlayers();i++){
      for(int j=0;j<baggPtr->getNumTypes(i);j++,k++){
        actions[k]=baggPtr->getNumActions(i,j);
      }
    }
    return actions;
  }

};


#endif

