#include "gnm.h"
#include "nfgame.h"
#include "makegame.h"
#include "BAGG/bagg.h"
#include "agentform.h"
#include "ubcsat-time.h"

#include <iostream>
using namespace std;

//GW gnm constants
#ifndef STEPS
  #define STEPS 100
#endif
#define FUZZ 1e-12
#ifndef LNMFREQ
  #define LNMFREQ 3
#endif
#define LNMMAX 10
#define LAMBDAMIN -0.2
#ifndef WOBBLE
  #define WOBBLE 0
#endif
#define THRESHOLD 1e-2

#define DEBUG 0
//#define FULLSTRAT 1
//#define POSPROB 0.4

//whether to use the full version of coffee shop game
#define FULL 1

#ifndef MAXEQL
  #define MAXEQL 1
#endif

#define DOAGG 1
#define DONF 0

void usage(char *name) {
  cout<<"usage:\n";
    //<< name <<" -sym numPlayers numActionNodes gameSeed strategySeed numRuns"
    //<<endl;
  cout<<name
    <<" -coffee[1|2]  numPlayers numTypes numRows numCols gameSeed strategySeed numRuns"
    <<endl;
  cout<<name
	  << " -f filename strategySeed numRuns"
	  <<endl;
}

int main(int argc, char **argv) {
  int S,T=0,r=0,c=0, argbase=0;
  enum {SYM, COFFEE,F} gtype; 
  projtype p=P_SUM;
  if (argc <3){
    usage(argv[0]);
    return -1;
  }
  int n;
  
/*
  if (strcmp(argv[1], "-sym")==0){ 
    if (argc <7){
      usage(argv[0]);
      return -1;
    }
    gtype=SYM;
    S = atoi(argv[3]);
    n = atoi(argv[2]);
  }
  else 
*/
  if (strcmp(argv[1], "-f")==0 ){
	  if(argc< 5){
		  usage(argv[0]);
		  return -1;
	  }
	  gtype=F;
	  argbase=-2;
  }
  else {
    if (argc<9) {usage(argv[0]); return -1;}
    gtype=COFFEE;
    if (strcmp(argv[1], "-coffee2")==0) p = P_EXIST;
    else p=P_SUM;
    n = atoi(argv[2]);
    T=atoi(argv[3]);
    r=atoi(argv[4]);
    c =atoi(argv[5]);
    argbase+=2; 
  }
  
  
  
  int gameSeed = atoi(argv[4+argbase]);
  int strategySeed = atoi(argv[5+argbase]);
  int numStrategies =atoi(argv[6+argbase]);

  
  srand48(gameSeed);
  //make the AGG 
  agentform *baggPtr=NULL;
  //if (gtype==SYM)
  //  baggPtr = new aggame(makeRandSymAGG (n,S,gameSeed));
  //else 
  if(gtype==COFFEE)
    baggPtr = new agentform(makeRandCoffeeShopBAGG(n,T,r,c,gameSeed,p));
  else 
    baggPtr = new agentform( bagg::makeBAGG(argv[2]) );
  
/*
  S=aggPtr->getNumActionNodes();
  n=aggPtr->getNumPlayers();
  if (DEBUG){
    for (int i=0;i<S; i++) {
      cout<<"AGG: payoffs for action node #"<<i<<endl;
      aggPtr-> printPayoffs(cout, i); 
    }
  }
*/
  //convert to NFgame
  nfgame *nfPtr=NULL;
  if(DONF){
    nfPtr =  makeNFGame (baggPtr);
    if(DEBUG){
      cout<<"Corresponding NFGame is: "<<endl;
      cout<< (*nfPtr);
    }
  }

  //vector<cvector> strats (numStrategies, baggPtr->getNumActions());
  cvector aggPayoff(numStrategies);
  cvector nfPayoff(numStrategies);
  srand48(strategySeed);
  double aggTime=0, nfTime=0, aggJTime,nfJTime,aggpayoffJTime;


  //GW
  cout<<"testing GW algorithm:"<<endl;
  srand48(strategySeed);
  int aggNumEq,nfNumEq;
  
  cvector g(baggPtr->getNumActions());
  
  cvector **nfAnswers;
  cvector **aggAnswers;
  double nfGWTime=0,aggGWTime=0, nfTotalGWTime=0,aggTotalGWTime=0;
  //cvector ratio (numStrategies);

  for (int k=0;k<numStrategies; ) {
    cout<< "run #"<<k<<endl;
    for(int i = 0; i < baggPtr->getNumActions(); i++) {
	g[i] = drand48();
    }
    g /= g.norm(); // normalized


    if (DOAGG){
      StartClock();
  
      //cvector g2=g;
      aggNumEq = GNM(*baggPtr, g, aggAnswers, STEPS, FUZZ, LNMFREQ, LNMMAX, LAMBDAMIN, WOBBLE, THRESHOLD,MAXEQL);

      for(int i = 0; i < aggNumEq; i++) {
        cout << *(aggAnswers[i]) << endl;
      
      }
      aggGWTime=TimeElapsed();
      if (aggNumEq>0){
	cout<<"Expected utility for each player under a sample eqlm:"<<endl;
	for (int player=0;player<baggPtr->getNumPlayers();player++){
	  cout<<baggPtr->getMixedPayoff(player, *(aggAnswers[0])) <<" ";
	}
	cout<<endl;
        //cout<<endl<<"Expected configuration of a sample eqlm:"<<endl;

        //cout <<baggPtr->getExpectedConfig(*(aggAnswers[0])) <<endl;
      }

      for(int i = 0; i < aggNumEq; i++) {
        delete aggAnswers[i];
      }
      free (aggAnswers);
    }
    
  
    if (DOAGG)cout<<"CPU Time for BAGG "<<aggGWTime<<endl;
    if (DOAGG&&aggNumEq==0) continue;

    aggTotalGWTime+=aggGWTime;
    ++k;
  } //end for(int k..
  cout<<endl<<"Avg CPU Time :"<<endl;
  cout<< aggTotalGWTime / numStrategies <<endl;

  delete baggPtr;
  return 0;
}
