#ifndef __PROJ_FUNC_H
#define __PROJ_FUNC_H

#include <set>
using namespace std;

//types of projection:
//sum, existence, highest, lowest
typedef enum{P_SUM,P_EXIST, P_HIGH,P_LOW} projtype;


//projection mappings
struct proj_func{
	int Default;
	proj_func(int def): Default(def) {}
	virtual ~proj_func() {}
	virtual int operator()(int x,int y)   =0;
	virtual int operator()(multiset<int>& s) =0;
};

struct proj_func_SUM: public proj_func {
    proj_func_SUM():proj_func(0) {}
    inline int operator() (int x,int y){return x+y;}
    inline int operator()(multiset<int>& s){return s.size();}
};
struct proj_func_EXIST: public proj_func{
    proj_func_EXIST() :proj_func(0) {}
    inline int operator() (int x, int y) {return (x+y>0);}
    inline int operator() (multiset<int>& s){ return (s.size()>0);}
};
struct proj_func_HIGH:public proj_func{
    proj_func_HIGH(int def): proj_func(def) {}
    inline int operator()(int x, int y){
	if(x==Default) return y;
	if (y==Default)return x;
	return ( (x>y)?x:y );
    }
    inline int operator()(multiset<int>& s){
	if (s.empty()) return Default;
	return *(s.rbegin());
    }
};
struct proj_func_LOW: public proj_func{
    proj_func_LOW(int def): proj_func(def) {}
    inline int operator()(int x, int y){
	if (x==Default) return y;
	if (y==Default )return x;
	return ( (x<y)?x:y);
    }
    inline int operator()(multiset<int>& s){
	if (s.empty() ) return Default;
	return *(s.begin());
    }
};	

inline proj_func* make_proj_func(projtype type, int Default){
  switch(type){
	case P_SUM: return (new proj_func_SUM);
	case P_EXIST: return (new proj_func_EXIST);
	case P_HIGH:  return (new proj_func_HIGH(Default));
	case P_LOW:     return (new proj_func_LOW(Default));
	default:    return NULL;
  }
}

#endif
