#include "gnm.h"
#include "nfgame.h"
#include "makegame.h"
#include "libagg/agg.h"
#include "aggame.h"
#include "ubcsat-time.h"

#include <iostream>
using namespace std;

//GW gnm constants
#ifndef STEPS
  #define STEPS 100
#endif
#define FUZZ 1e-12
#ifndef LNMFREQ
  #define LNMFREQ 3
#endif
#define LNMMAX 10
#define LAMBDAMIN -0.2
#ifndef WOBBLE
  #define WOBBLE 0
#endif
#define THRESHOLD 1e-2

#define FULLSTRAT 1
#define POSPROB 0.4
#define DEBUG 0
//whether to use the full version of coffee shop game
#define FULL 1

#ifndef MAXEQL
  #define MAXEQL 1
#endif

#define DOAGG 1
#define DONF 1
#define DOPAYOFF 1
#define DOJAC 0
#define DOSYM 0
#define DOGW 0

void usage(char *name) {
  cout<<"usage:\n"<< name 
    <<" -sym numPlayers numActionNodes gameSeed strategySeed numStrategies"
    <<endl;
  cout<<name
    <<" -coffee[1|2]  numPlayers numRows numCols gameSeed strategySeed numStrategies"
    <<endl;
  cout<<name
	  << " -f filename strategySeed numStrategies"
	  <<endl;
}

void randSymStrats(gnmgame* g, vector<cvector>& strats, vector<cvector>& sym_strats,  int numStrats,int seed, bool full_sup, double posprob ){
  srand48(seed);
  for(int i=0;i<numStrats; i++){

    if(full_sup){
      g->randomFullStrategy(strats[i], 0);
    } else {
      g->randomSupportStrategy(strats[i], 0, posprob);
    }
    for(int j=g->firstAction(0);j<g->lastAction(0);j++){
	sym_strats[i][j] = strats[i][j];
    }
    for (int player=1;player<g->getNumPlayers();player++){
        for(int j=0; j<g->getNumActions(player); j++){
          assert(g->getNumActions(player)==g->getNumActions(0));
          strats[i][j+g->firstAction(player)]=sym_strats[i][j];
        } 
    }
  }

}

int main(int argc, char **argv) {
  int S,r=0,c=0, argbase=0;
  enum {SYM, COFFEE,F} gtype; 
  projtype p=P_SUM;
  if (argc <3){
    usage(argv[0]);
    return -1;
  }
  int n;
  
  if (strcmp(argv[1], "-sym")==0){ 
    if (argc <7){
      usage(argv[0]);
      return -1;
    }
    gtype=SYM;
    S = atoi(argv[3]);
    n = atoi(argv[2]);
  }
  else if (strcmp(argv[1], "-f")==0 ){
	  if(argc< 5){
		  usage(argv[0]);
		  return -1;
	  }
	  gtype=F;
	  argbase=-2;
  }
  else {
    if (argc<8) {usage(argv[0]); return -1;}
    gtype=COFFEE;
    if (strcmp(argv[1], "-coffee2")==0) p = P_EXIST;
    else p=P_SUM;
    n = atoi(argv[2]);
    r=atoi(argv[3]);
    c=atoi(argv[4]);
    S =r*c+1;
    argbase++; 
  }
  
  
  
  int gameSeed = atoi(argv[4+argbase]);
  int strategySeed = atoi(argv[5+argbase]);
  int numStrategies =atoi(argv[6+argbase]);

  
  srand48(gameSeed);
  //make the AGG 
  aggame *aggPtr=NULL;
  if (gtype==SYM)
    aggPtr = new aggame( makeRandSymAGG (n,S,gameSeed) );
  else if(gtype==COFFEE)
    aggPtr = new aggame( makeRandCoffeeShopAGG(n,r,c,gameSeed,p, FULL) );
  else 
    aggPtr = new aggame( agg::makeAGG(argv[2]) );
  

  S=aggPtr->getNumActionNodes();
  n=aggPtr->getNumPlayers();



  if (DEBUG){
    for (int i=0;i<S; i++) {
      cout<<"AGG: payoffs for action node #"<<i<<endl;
      aggPtr->aggPtr->printPayoffs(cout, i); 
    }
  }
  //convert to NFgame
  nfgame *nfPtr=NULL;
  if(DONF){
    nfPtr =  makeNFGame (aggPtr);
    if(DEBUG){
      cout<<"Corresponding NFGame is: "<<endl;
      cout<< (*nfPtr);
    }
  }

  vector<cvector> strats (numStrategies, aggPtr->getNumActions());
  cvector aggPayoff(numStrategies);
  cvector nfPayoff(numStrategies);
  srand48(strategySeed);
  double aggTime=0, nfTime=0, aggJTime,nfJTime,aggpayoffJTime;

  cerr<< "Full Support Strategies? "<<FULLSTRAT<<endl; 
  for (int i=0;i<numStrategies; i++){
    
    if(FULLSTRAT)aggPtr->randomFullStrategy(strats[i]);
    else aggPtr->randomSupportStrategy(strats[i],POSPROB);
    if(DEBUG)cout<<"Strategy profile #"<<i<<": \n"<<strats[i]<<endl;
  }
  vector<cvector> nfstrats (strats);


  // COMPUTE EXPECTED PAYOFFS
if(DOPAYOFF){
 if (DONF){
  
  for (int i=0;i<numStrategies; i++){
    for (int j=0;j<aggPtr->getNumActions(n-1);j++)nfstrats[i][nfPtr->firstAction(n-1)+j]=0;
    nfstrats[i][nfPtr->firstAction(n-1) + i%aggPtr->getNumActions(n-1)]=1.0;
  }

  StartClock();
  for (int i=0;i<numStrategies; i++){
            nfPayoff[i] = nfPtr->getMixedPayoff(n-1,nfstrats[i]);
  }            
  
  nfTime = TimeElapsed();
 } //end if (DONF)

 if(DOAGG){
  StartClock();
  for (int i=0;i<numStrategies; i++)
      aggPayoff[i] = aggPtr->aggPtr->getV(n-1, i%aggPtr->getNumActions(n-1),  strats[i]);
  aggTime= TimeElapsed();
 } //end if(DOAGG)   
  
  if(DOAGG)cout<< "CPU Time for computing AGG payoff: "<<aggTime<<endl;
  if (DONF)cout<<"CPU Time for computing NFG payoff: "<<nfTime<<endl;
  
  cout.flush();
  if(DONF&&DOAGG){
    assert ( (aggPayoff-nfPayoff).norm() <= FUZZ );
  }
  if (DOAGG) cout<<aggTime<<" ";
  if (DONF) cout<<nfTime;
  cout<<endl<<endl;
  if(DEBUG&&DOAGG&&DONF && strats.size()>0) {
      cout<<"Testing getPayoffVector:\n";
      for (int pl=0;pl<aggPtr->getNumPlayers();++pl){
        cvector aggvec (aggPtr->getNumActions(pl));
        cvector nfvec (aggPtr->getNumActions(pl));
        aggPtr->getPayoffVector(aggvec,pl, strats[0]);
	nfPtr->getPayoffVector(nfvec, pl, strats[0]);
	assert( (aggvec-nfvec).norm()<=FUZZ);
      }
      cout<<"Finished testing getPayoffVector."<<endl<<endl;
  }
} //end if(DOPAYOFF)

  if (gtype==SYM && DOSYM){ 
    double nfSymTime, aggSymTime, aggSymGrayTime;
    cvector nfSymPay(numStrategies), aggSymPay(numStrategies), aggSymGrayPay(numStrategies);
    vector<cvector> s (numStrategies, S);
    vector<cvector> full_s (numStrategies, S*n );
    randSymStrats(aggPtr, full_s,s, numStrategies, strategySeed,FULLSTRAT,POSPROB);
    cout<<endl<<"Testing symmetric exp. payoff"<<endl;
    //nf
    cout<<"normal form: ";
    StartClock();
    for (int i=0;i< numStrategies; i++)
      nfSymPay[i] = nfPtr->getMixedPayoff(0, full_s[i]);
    nfSymTime=TimeElapsed();
    cout<<" Time: "<<nfSymTime<<endl;
    //agg
    StartClock();
    for(int i=0; i<numStrategies; i++)
      aggSymPay[i]=aggPtr->getMixedPayoff(0,full_s[i]);
    aggSymTime=TimeElapsed();
    cout<<"agg Time: "<<aggSymTime<<endl;
    //agg/graycode
    StartClock();
    for (int i=0;i<numStrategies;i++)
      aggSymGrayPay[i]= aggPtr->aggPtr->getSymMixedPayoff( s[i]);
    aggSymGrayTime=TimeElapsed();
    cout<<"agg/graycode Time: "<<aggSymGrayTime<<endl;

    //assert( (aggSymGrayPay-nfSymPay).norm() <=FUZZ);
    for (int i=0;i<numStrategies; i++){
      if(fabs(aggSymGrayPay[i]-nfSymPay[i])>FUZZ){
	cout<<"Error: different results."<<endl;
	cout<<"Strategy: "<<s[i] <<endl;
	cout <<"Results: agg/gray: "<<aggSymGrayPay[i]<<" nf: "
		<<nfSymPay[i] <<endl;
	exit(1);
      }
    }
  }

  //testing payoffMatrix
if(DOJAC){
  //cout<<"computing jacobian for strat "<<endl<<strats[0]<<endl;
  cmatrix J (aggPtr->getNumActions(),aggPtr->getNumActions());
  cmatrix J1 (aggPtr->getNumActions(),aggPtr->getNumActions());
  cmatrix J2(aggPtr->getNumActions(),aggPtr->getNumActions());

  StartClock();
  if (DOAGG)for (int i=0;i<numStrategies; i++){
    aggPtr->payoffMatrix(J1, strats[i],FUZZ);
  }
  aggJTime=TimeElapsed();

  StartClock();
  double fuzzcount,fuzz=FUZZ;
  if(DOAGG)for (int i=0; i<numStrategies; i++){
    //diagonal
    for (int rown=0;rown<n;rown++){
	  fuzzcount=fuzz;
	  for (int rowi=aggPtr->firstAction(rown); rowi<aggPtr->lastAction(rown);rowi++){
	    for (int coli=aggPtr->firstAction(rown);coli<aggPtr->lastAction(rown);coli++){
		J[rowi][coli]=fuzzcount;
		fuzzcount+=fuzz;
	    }
	  }
    }
    for (int rown=0;rown<n;rown++){
      for (int coln=0;coln<n;coln++)if(rown!=coln){
        for(int act1=0;act1<aggPtr->getNumActions(rown);act1++){
	  for (int act2=0;act2<aggPtr->getNumActions(coln);act2++){
            J[act1+aggPtr->firstAction(rown)][act2+aggPtr->firstAction(coln)]=
		aggPtr->aggPtr->getJ(rown,act1,coln,act2,strats[i]);
          }
        }
      }
    }

  }

  aggpayoffJTime=TimeElapsed();

  StartClock();
  if (DONF)for (int i=0;i<numStrategies; i++){
    nfPtr->payoffMatrix(J2,strats[i],FUZZ);
  }
  nfJTime=TimeElapsed();

  if(DOAGG)cout<<endl<<"max diff J (AGG payoffs) and J1(AGG Jacobian):"<<(J1-J).absmax()<<endl;
  if(DOAGG&&DONF){
      cout<<endl<<"max diff J1 (AGG Jacobian) and J2 (NF):"<<(J1-J2).absmax()<<endl;
	  //<<"diff matrix:\n"<<J1-J2<<endl;
      for (int rown=0;rown<n;rown++){
       for (int coln=0;coln<n;coln++){
        for(int act1=0;act1<aggPtr->getNumActions(rown);act1++){
	  for (int act2=0;act2<aggPtr->getNumActions(coln);act2++){
	    double d = fabs(J1[act1+aggPtr->firstAction(rown)][act2+aggPtr->firstAction(coln)]-J2[act1+aggPtr->firstAction(rown)][act2+aggPtr->firstAction(coln)]);
	    if (d>FUZZ){
		cout<<"ERROR: dif results at Jac for player "<<rown<<" act "<<act1<<" player "<<coln<<" act "<<act2<<": diff "<<d<<endl;
	    }
	  }
	}
       }
      }
      cout<< "max diff J (AGG payoffs) and J2 (NF):"<<(J-J2).absmax()<<endl;
  }
  
/*
  if (DEBUG) {
    cout<< "jacobian by agg: "<<endl<<J1<<endl;
    cout<<"jacobian by nf: "<<endl<<J2<<endl;
    cout<<endl<<"The diff: "<<endl<<J1-J2<<endl;
  }
*/
  cout<<"CPU Time for computing AGG Jacobian "<<aggJTime<<endl;
  cout<<"CPU Time for Jacobian using AGG payoffs "<<aggpayoffJTime<<endl; 
  cout<<"CPU Time for computing NFG Jacobian "<<nfJTime<<endl;
}

  //GW
  if (!DOGW) {
    delete aggPtr; 
    if(nfPtr) delete nfPtr; 
    return 0;
  }
  cerr<<"testing GW algorithm:"<<endl;
  srand48(strategySeed);
  int aggNumEq,nfNumEq;
  
  cvector g(aggPtr->getNumActions());
  
  cvector **nfAnswers;
  cvector **aggAnswers;
  double nfGWTime=0,aggGWTime=0, nfTotalGWTime=0,aggTotalGWTime=0;
  cvector ratio (numStrategies);

  for (int k=0;k<numStrategies; ) {
    cerr<< "run #"<<k<<endl;
    for(int i = 0; i < aggPtr->getNumActions(); i++) {
	g[i] = drand48();
    }
    g /= g.norm(); // normalized


    if (DONF) {
      cerr<<endl<<"NFG:"<<endl;

      StartClock();
  
      cvector g1 =g;
      nfNumEq = GNM(*nfPtr, g1, nfAnswers, STEPS, FUZZ, LNMFREQ, LNMMAX, LAMBDAMIN, WOBBLE, THRESHOLD,MAXEQL);
 
      for(int i = 0; i < nfNumEq; i++) {
        cerr << *(nfAnswers[i]) << endl;
      
      }

    
      nfGWTime=TimeElapsed();
    }
    if (DOAGG){
      cerr<<endl<<"AGG"<<endl;
      StartClock();
  
      cvector g2=g;
      aggNumEq = GNM(*aggPtr, g2, aggAnswers, STEPS, FUZZ, LNMFREQ, LNMMAX, LAMBDAMIN, WOBBLE, THRESHOLD,MAXEQL);

      for(int i = 0; i < aggNumEq; i++) {
        cerr << *(aggAnswers[i]) << endl;
      
      }
      aggGWTime=TimeElapsed();
      if (aggNumEq>0){
	cerr<<"Expected utility for each player under a sample eqlm:"<<endl;
	for (int player=0;player<aggPtr->getNumPlayers();player++){
	  cerr<<aggPtr->getMixedPayoff(player, *(aggAnswers[0])) <<" ";
	}
        cerr<<endl<<"Expected configuration of a sample eqlm:"<<endl;

        cerr <<aggPtr->getExpectedConfig(*(aggAnswers[0])) <<endl;
      }    

    }
  
    if (DOAGG)cerr<<"CPU Time for AGG "<<aggGWTime<<"\t";
    if (DONF)cerr<<"CPU Time for NFG "<<nfGWTime<<endl;
    if(DOAGG&&DONF &&nfNumEq!=aggNumEq){
	cerr<<"WARNING: numbers of equilibria do not match"<<endl;
        continue;
    }
    if (DOAGG&&aggNumEq==0) continue;

    if (DEBUG&&DOAGG&&DONF)for(int i=0;i<aggNumEq;i++){
      cout<<"norm of diff for solution #"<<i<<": "<<(*(aggAnswers[i])- *(nfAnswers[i])).norm()<<endl;
    }

    if(DOAGG)for(int i = 0; i < aggNumEq; i++) {
      delete aggAnswers[i];
    }
    if(DONF)for (int i=0;i<nfNumEq;i++){
      delete nfAnswers[i];
    }
    if(DOAGG) free (aggAnswers);
    if(DONF) free (nfAnswers);
    nfTotalGWTime+=nfGWTime;
    aggTotalGWTime+=aggGWTime;
    ratio[k] = nfGWTime /aggGWTime;
    cout<<nfGWTime<<"\t"<<aggGWTime<<"\t"<<ratio[k]<<endl;
    ++k;
  } //end for(int k..
  cout<<"Avg CPU Time :"<<endl;
  if(DOAGG)cout<<"AGG: "<< aggTotalGWTime / numStrategies <<endl;
  if(DONF)cout<<"NFG: "<<nfTotalGWTime/ numStrategies <<endl;
  if(DOAGG&&DONF)cout<< "ratio of NFG time / AGG time: "<< ratio<<endl;

  cout<<aggTotalGWTime / numStrategies<<" "
      <<nfTotalGWTime/ numStrategies <<" "
      <<ratio <<endl;
  delete aggPtr;
  if(nfPtr) delete nfPtr;
  
  return 0;
}
